/**
 * \file filesrenamer.cpp
 * Rename files dialog.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 8 Dec 2009
 *
 * Copyright (C) 2009  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "filesrenamer.h"
#include "application.h"
#include "tagger.h"
#include "tagrender.h"
#include "legendbutton.h"
#ifdef HAVE_QT5
#include <QtWidgets>
#else
#include <QtGui>
#endif
#include "highlightdelegate.h"
#include "filedelegatehighlighter.h"
#include "fancylineedit.h"
#include "model.h"

FileRenameDialog::FileRenameDialog(Model *model, const QStringList &fileNames,  QWidget *parent)
    : QDialog(parent), m(model), oldFileNames(fileNames)
{
    setWindowTitle(tr("Move/Copy/Rename files"));
    if (!m || !m->hasSelection()) return;

    filenameRenderer = new FileNameRenderer(m);
    highlighter = new FileDelegateHighlighter();

    ////////////////////////////////////////////////////////////////////////////
    QGroupBox *whatBox = new QGroupBox(tr("Do what:"),this);

    operationComboBox = new QComboBox(this);
    operationComboBox->addItem(tr("rename/move files"));
    operationComboBox->addItem(tr("copy files"));
    operationComboBox->setEditable(false);
    operationComboBox->setCurrentIndex(App->renamingOperation);
    connect(operationComboBox,SIGNAL(currentIndexChanged(int)),SLOT(handleRemoveFolders()));

    removeFolderCheckBox = new QCheckBox(tr("Delete empty folder after renaming"),this);
    removeFolderCheckBox->setChecked(App->removeFolder);
    connect(removeFolderCheckBox,SIGNAL(toggled(bool)),SLOT(handleRemoveFolders()));


    QHBoxLayout *whatLayout = new QHBoxLayout;
    whatLayout->addWidget(operationComboBox);
    whatLayout->addWidget(removeFolderCheckBox);
    whatLayout->addStretch();
    whatBox->setLayout(whatLayout);

    ////////////////////////////////////////////////////////////////////////////
    directoryGroup = new QGroupBox(tr("Output folder"),this);
    directoryGroup->setCheckable(true);
    connect(directoryGroup,SIGNAL(clicked()),SLOT(directoryChanged()));

    QString destinationFolder = App->destinationFolder;
    directoryEdit = new FancyLineEdit(this);
    QPixmap pixmap(16, 16);
    pixmap.fill(Qt::transparent);
    QPainter painter(&pixmap);
    painter.drawText(0,14,QSL("..."));

    directoryEdit->setButtonPixmap(FancyLineEdit::Right, pixmap);
    directoryEdit->setButtonVisible(FancyLineEdit::Right, true);
    directoryEdit->setButtonToolTip(FancyLineEdit::Right, tr("Open existing folder"));
    directoryEdit->setAutoHideButton(FancyLineEdit::Right, false);
    connect(directoryEdit, SIGNAL(rightButtonClicked()), this, SLOT(setExistingDirectory()));

    directoryEdit->setText(destinationFolder);
    connect(directoryEdit,SIGNAL(textChanged(QString)),this,SLOT(directoryChanged()));
    directoryGroup->setChecked(!destinationFolder.isEmpty());

    QHBoxLayout *dirLayout = new QHBoxLayout;
    dirLayout->addWidget(directoryEdit);
    directoryGroup->setLayout(dirLayout);

    ////////////////////////////////////////////////////////////////////////////
    QGroupBox *patternBox = new QGroupBox(tr("Output file name pattern"),this);

    patternEdit = new QComboBox(this);
    patternEdit->setEditable(true);
    patternEdit->setAutoCompletion(true);
    patternEdit->setAutoCompletionCaseSensitivity(Qt::CaseSensitive);
    if (!App->patterns.isEmpty()) {
        patternEdit->insertItems(0,App->patterns);
        filenameRenderer->setPattern(App->patterns.first());
        patternEdit->setEditText(App->patterns.first());
    }
    connect(patternEdit,SIGNAL(currentIndexChanged(QString)),this,SLOT(patternChanged(QString)));
    connect(patternEdit,SIGNAL(editTextChanged(QString)),this,SLOT(patternChanged(QString)));

    legendButton = new LegendButton(this);
    legendButton->setCategories(LegendButton::WritablePlaceholders | LegendButton::ReadOnlyPlaceholders);
    legendButton->retranslateUi();
    connect(legendButton,SIGNAL(placeholderChosen(QString)),SLOT(insertLegend(QString)));


    QHBoxLayout *patternLayout = new QHBoxLayout;
    patternLayout->addWidget(patternEdit,1);
    patternLayout->addWidget(legendButton);
    patternBox->setLayout(patternLayout);

    ////////////////////////////////////////////////////////////////////////////


    charactersCheckBox = new QCheckBox(tr("Replace special Windows characters \\/:*?\"<>| with"),this);
    charactersCheckBox->setToolTip(tr("Characters \\/:*| will be replaced by this text\n"
                                      "character ? will be removed\n"
                                      "characters <> will be replaced by (),\n"
                                      "and \" by two single quotes ''"));
    charactersCheckBox->setChecked(App->replaceWinChars);
    connect(charactersCheckBox,SIGNAL(stateChanged(int)),SLOT(replaceWinCharsChanged()));
    replaceCharacterEdit = new QLineEdit(this);
    replaceCharacterEdit->setText(App->winCharsReplacer);
    replaceCharacterEdit->setEnabled(App->replaceWinChars);
    connect(replaceCharacterEdit,SIGNAL(textEdited(QString)),SLOT(replaceWinCharsChanged()));

    spacesCheckBox = new QCheckBox(tr("Replace spaces with"));
    spacesCheckBox->setChecked(App->replaceSpaces);
    connect(spacesCheckBox,SIGNAL(stateChanged(int)),SLOT(replaceSpacesChanged()));
    spacesEdit = new QLineEdit(this);
    spacesEdit->setText(App->spacesReplacer);
    spacesEdit->setEnabled(App->replaceSpaces);
    connect(spacesEdit,SIGNAL(textEdited(QString)),SLOT(replaceSpacesChanged()));

    fileLengthSpinBox = new QSpinBox(this);
    fileLengthSpinBox->setMinimum(5);
    fileLengthSpinBox->setMaximum(25000);
    fileLengthSpinBox->setValue(App->maximumLength);
    fileLengthSpinBox->setEnabled(App->trimFileLength);
    fileLengthCheckBox = new QCheckBox(tr("Maximum file name length"),this);
    fileLengthCheckBox->setChecked(App->trimFileLength);
    connect(fileLengthSpinBox,SIGNAL(valueChanged(int)),SLOT(fileLengthChanged(int)));
    connect(fileLengthCheckBox,SIGNAL(toggled(bool)),SLOT(updateFileLength(bool)));



    caseCheckBox = new QCheckBox(tr("Case"),this);
    caseCheckBox->setChecked(App->changeCase);
    connect(caseCheckBox,SIGNAL(toggled(bool)),SLOT(updateCase(bool)));

    caseComboBox = new QComboBox(this);
    caseComboBox->setEditable(false);
    caseComboBox->addItem(tr("lower"));
    caseComboBox->addItem(tr("UPPER"));
    caseComboBox->addItem(tr("Only first up"));
    caseComboBox->addItem(tr("Every First Up"));
    caseComboBox->setCurrentIndex(App->ccase);
    caseComboBox->setEnabled(App->changeCase);
    connect(caseComboBox,SIGNAL(currentIndexChanged(int)),SLOT(caseChanged(int)));

    diacriticsCheckBox = new QCheckBox(tr("Remove diacritics"),this);
    diacriticsCheckBox->setChecked(App->removeDiacritics);
    connect(diacriticsCheckBox,SIGNAL(stateChanged(int)),SLOT(diacriticsChanged()));

    applyToFoldersCheckBox = new QCheckBox(tr("Apply these options to created folders as well"),this);
    applyToFoldersCheckBox->setChecked(App->applyToFolders);
    connect(applyToFoldersCheckBox,SIGNAL(stateChanged(int)),SLOT(applyToFoldersChanged()));


    QGridLayout *groupBox2Layout = new QGridLayout;
    groupBox2Layout->addWidget(charactersCheckBox,0,0);
    groupBox2Layout->addWidget(replaceCharacterEdit,0,1);
    groupBox2Layout->addWidget(spacesCheckBox,1,0);
    groupBox2Layout->addWidget(spacesEdit,1,1);
    groupBox2Layout->addWidget(fileLengthCheckBox,2,0);
    groupBox2Layout->addWidget(fileLengthSpinBox,2,1);
    groupBox2Layout->addWidget(caseCheckBox,0,3);
    groupBox2Layout->addWidget(caseComboBox,0,4);
    groupBox2Layout->addWidget(diacriticsCheckBox,1,3,1,2);
    groupBox2Layout->addWidget(applyToFoldersCheckBox,2,3,1,2);
    groupBox2Layout->setColumnMinimumWidth(2,40);
    groupBox2Layout->setColumnStretch(0,1);
    groupBox2Layout->setColumnStretch(3,1);
    groupBox2Layout->setColumnStretch(4,1);


    ////////////////////////////////////////////////////////////////////////////
    table = new QTableWidget(oldFileNames.count(),2,this);
    table->setHorizontalHeaderLabels(QStringList() << tr("Old names") << tr("New names"));
    table->horizontalHeader()->setStretchLastSection(true);
    table->horizontalHeader()->resizeSection(0, 350);
    table->setTextElideMode(Qt::ElideLeft);
    HighlightDelegate *delegate = new HighlightDelegate(this);
    delegate->setHighlighter(highlighter);
    table->setItemDelegateForColumn(1, delegate);
    for (int i=0; i<oldFileNames.count(); ++i) {
        QString s = QFileInfo(oldFileNames.at(i)).fileName();
        QTableWidgetItem *item = new QTableWidgetItem(s);
        item->setFlags(Qt::ItemIsSelectable  | Qt::ItemIsEnabled);
        table->setItem(i,0,item);
        QTableWidgetItem *item1 = new QTableWidgetItem(oldFileNames.at(i));
        item1->setFlags(Qt::ItemIsSelectable  | Qt::ItemIsEnabled | Qt::ItemIsEditable);
        table->setItem(i,1,item1);
    }

    ////////////////////////////////////////////////////////////////////////////
    QDialogButtonBox *buttonBox = new QDialogButtonBox(Qt::Horizontal);
    runButton = buttonBox->addButton(tr("Run"), QDialogButtonBox::AcceptRole);
    buttonBox->addButton(QDialogButtonBox::Cancel);
    connect(buttonBox, SIGNAL(accepted()), this, SLOT(run()));
    connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));
    //////////////////////////////////////////////////////////////////////////////

    QVBoxLayout *layout = new QVBoxLayout;
    layout->addWidget(whatBox);
    layout->addWidget(directoryGroup);
    layout->addWidget(patternBox);
    layout->addLayout(groupBox2Layout);
    layout->addWidget(table);
    layout->addWidget(buttonBox);
    setLayout(layout);
    resize(800,550);

    setTabOrder(operationComboBox,removeFolderCheckBox);
    setTabOrder(removeFolderCheckBox,directoryGroup);
    setTabOrder(directoryGroup,directoryEdit);
    setTabOrder(directoryEdit,patternEdit);
    setTabOrder(patternEdit,legendButton);
    setTabOrder(legendButton,charactersCheckBox);
    setTabOrder(charactersCheckBox,replaceCharacterEdit);
    setTabOrder(replaceCharacterEdit,spacesCheckBox);
    setTabOrder(spacesCheckBox,spacesEdit);
    setTabOrder(spacesEdit,fileLengthCheckBox);
    setTabOrder(fileLengthCheckBox,fileLengthSpinBox);
    setTabOrder(fileLengthSpinBox,caseCheckBox);
    setTabOrder(caseCheckBox,caseComboBox);
    setTabOrder(caseComboBox,diacriticsCheckBox);
    setTabOrder(diacriticsCheckBox,buttonBox);
    setTabOrder(buttonBox,operationComboBox);

    table->setFocusPolicy(Qt::NoFocus);
    updateTable();
}

FileRenameDialog::~FileRenameDialog()
{
    delete filenameRenderer;
    delete highlighter;
}

void FileRenameDialog::patternChanged(const QString &pattern)
{
    filenameRenderer->setPattern(pattern);
    updateTable();
}

void FileRenameDialog::directoryChanged()
{
    if (directoryGroup->isChecked())
        App->destinationFolder = directoryEdit->text();
    else
        App->destinationFolder.clear();
    updateTable();
}

void FileRenameDialog::replaceWinCharsChanged()
{
    App->replaceWinChars = charactersCheckBox->isChecked();
    App->winCharsReplacer = replaceCharacterEdit->text();
    replaceCharacterEdit->setEnabled(charactersCheckBox->isChecked());
    updateTable();
}

void FileRenameDialog::replaceSpacesChanged()
{
    App->replaceSpaces = spacesCheckBox->isChecked();
    App->spacesReplacer = spacesEdit->text();
    spacesEdit->setEnabled(spacesCheckBox->isChecked());
    updateTable();
}

void FileRenameDialog::diacriticsChanged()
{
    App->removeDiacritics = diacriticsCheckBox->isChecked();
    updateTable();
}

void FileRenameDialog::caseChanged(const int value)
{
    App->ccase = value;
    updateTable();
}

void FileRenameDialog::updateCase(bool b)
{
    caseComboBox->setEnabled(b);
    App->changeCase = b;
    updateTable();
}

void FileRenameDialog::fileLengthChanged(int value)
{
    App->maximumLength = value;
    updateTable();
}

void FileRenameDialog::updateFileLength(bool value)
{
    fileLengthSpinBox->setEnabled(value);
    App->trimFileLength = value;
    updateTable();
}

void FileRenameDialog::applyToFoldersChanged()
{
    App->applyToFolders = applyToFoldersCheckBox->isChecked();
    updateTable();
}

void FileRenameDialog::updateTable()
{
    QStringList newFileNames = filenameRenderer->getNewTags();
    QStringList prefixes = filenameRenderer->getPrefixes();
    Q_ASSERT(newFileNames.size()==prefixes.size());
    for (int i=0; i<newFileNames.size(); ++i) {
        table->item(i,1)->setText(newFileNames.at(i));
        table->item(i,1)->setData(Qt::UserRole+1,prefixes.at(i));
    }
    runButton->setEnabled(newFileNames!=oldFileNames);
    runButton->setToolTip(newFileNames!=oldFileNames ? "":tr("Nothing to rename, file names not changed."));
}

void FileRenameDialog::handleRemoveFolders()
{
    removeFolderCheckBox->setEnabled(operationComboBox->currentIndex()==0);
    App->removeFolder = removeFolderCheckBox->isChecked();
    App->renamingOperation = operationComboBox->currentIndex();
}

void FileRenameDialog::insertLegend(const QString &s)
{
    patternEdit->lineEdit()->insert(s);
}


void FileRenameDialog::run()
{
    //add pattern to file renaming patterns
    App->addRenamePattern(patternEdit->currentText());

    newFileNames.clear();
    for (int i=0; i<oldFileNames.count(); ++i) newFileNames << table->item(i,1)->text();
    accept();
}

void FileRenameDialog::setExistingDirectory()
{
    QString directory = QFileDialog::getExistingDirectory(this,
            tr("Folder to move/copy files"),
            oldFileNames.isEmpty() ? QDir::homePath() : QDir(oldFileNames.at(0)).canonicalPath(),
            QFileDialog::DontResolveSymlinks | QFileDialog::ShowDirsOnly);
    if (!directory.isEmpty())
        directoryEdit->setText(directory);
}

