/**
 * \file tagger.cpp
 * Class for reading, writing and storing tags.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 8 Dec 2009
 *
 * Copyright (C) 2009  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "tagger.h"
#include <QFileInfo>
#include <QDir>
#include <QtDebug>
#include "enums.h"

#include <QStringList>

void append(QString &tagString, const QString &newTag, const QString &sep)
{

    //additional processing - appends new non-empty tags
    if (newTag.isEmpty()) return;

    if (tagString.isEmpty()) {
        tagString=newTag;
        return;
    }
    QStringList v = tagString.split(sep);
    if (!v.contains(newTag))
        tagString.append(sep+newTag);
}

void ReplayGainInfo::clear()
{
    albumPeak.clear();
    trackPeak.clear();
    albumGain.clear();
    trackGain.clear();
    albumMinMax.clear();
    trackMinMax.clear();
    loudness.clear();
    undo.clear();
}

bool ReplayGainInfo::isEmpty() const
{
    return albumPeak.isEmpty()
            && trackPeak.isEmpty()
            && albumGain.isEmpty()
            && trackGain.isEmpty()
            && albumMinMax.isEmpty()
            && trackMinMax.isEmpty()
            && loudness.isEmpty()
            && undo.isEmpty();
}

TagData::TagData() :
    length(0),
    sampleRate(0),
    channels(0),
    wasChanged(false),
    filetype(-1)
{}

TagData::TagData(const TagData &other) :
    QSharedData(other),
    icon(other.icon),
    fullFileName(other.fullFileName),
    fileNameExt(other.fileNameExt),
    bitrate(other.bitrate),
    //formattedLength(other.formattedLength),
    length(other.length),
    sampleRate(other.sampleRate),
    channels(other.channels),
    wasChanged(other.wasChanged),
    filetype(other.filetype),
    album(other.album),
    artist(other.artist),
    title(other.title),
    performer(other.performer),
    comment(other.comment),
    genre(other.genre),
    composer(other.composer),
    year(other.year),
    tracknumber(other.tracknumber),
    totalTracks(other.totalTracks),
    conductor(other.conductor),
    orchestra(other.orchestra),
    subtitle(other.subtitle),
    key(other.key),
    albumArtist(other.albumArtist),
    category(other.category),
    publisher(other.publisher),
    copyright(other.copyright),
    mood(other.mood),
    tempo(other.tempo),
    lyricist(other.lyricist),
    lyrics(other.lyrics),
    discnumber(other.discnumber),
    totaldiscs(other.totaldiscs),
    encodedby(other.encodedby),
    remixedby(other.remixedby),
    rating(other.rating),
    originalalbum(other.originalalbum),
    originalartist(other.originalartist),
    originallyricist(other.originallyricist),
    url(other.url),
    isrc(other.isrc),
    replayGainInfo(other.replayGainInfo),
    image(other.image),
    otherTags(other.otherTags)
{}

Tag::Tag()
{
    d = new TagData;
}

Tag::Tag(const QString &filename)
{
    d = new TagData;
    setFile(filename);
}

QString Tag::tag(const int &tagID) const
{
    switch (tagID) {
    case COMPOSER:    return d->composer; break;
    case ALBUM:       return d->album; break;
    case TITLE:       return d->title; break;
    case PERFORMER:   return d->performer; break;
    case ARTIST:      return d->artist; break;
    case GENRE:       return d->genre; break;
    case YEAR:        return d->year; break;
    case COMMENT:     return d->comment; break;
    case TRACKNUMBER: return d->tracknumber; break;
    case TOTALTRACKS: return d->totalTracks; break;
    case CONDUCTOR:   return d->conductor; break;
    case ORCHESTRA:   return d->orchestra; break;
    case SUBTITLE:    return d->subtitle; break;
    case KEY:         return d->key; break;
    case ALBUMARTIST: return d->albumArtist; break;

    case CATEGORY:   return d->category; break;
    case PUBLISHER:  return d->publisher; break;
    case COPYRIGHT:  return d->copyright; break;
    case MOOD:       return d->mood; break;
    case TEMPO:      return d->tempo; break;
    case LYRICIST:   return d->lyricist; break;
    case LYRICS:     return d->lyrics; break;
    case DISCNUMBER: return d->discnumber; break;
    case TOTALDISCS: return d->totaldiscs; break;
    case ENCODEDBY:  return d->encodedby; break;
    case REMIXEDBY:  return d->remixedby; break;
    case RATING:     return d->rating; break;
    case ORIGINALALBUM: return d->originalalbum; break;
    case ORIGINALARTIST: return d->originalartist; break;
    case ORIGINALLYRICIST: return d->originallyricist; break;
    case URL :        return d->url; break;
    case ISRC :       return d->isrc; break;
    }
    return QString();
}

QString Tag::nonEmptyTag(const int &tagID,const QString &defaultVal) const
{
    QString s=tag(tagID);
    if (s.isEmpty()) return defaultVal;
    return s;
}

QString Tag::userTag(const QString &key) const
{
    return d->otherTags.value(key);
}

QStringList Tag::userTagsKeys() const
{
    return d->otherTags.keys();
}

void Tag::setFile(const QString &fileName)
{
    d->fullFileName=fileName;
    QFileInfo fi(d->fullFileName);
    d->fileNameExt=fi.fileName();
    d->fileName = fi.completeBaseName();
    d->fileExt = fi.suffix();
    d->filePath = fi.canonicalPath();

    QString extension = d->fileExt.toLower();

    if (extension=="asf") extension=QSL("wma");
    else if (extension=="m4b" || extension=="m4p" || extension=="mp4"
             || extension=="3g2") extension=QSL("m4a");
    else if (extension=="aif") extension=QSL("aiff");

    QStringList list=QString("mp3.ogg.wma.flac.m4a.mpc.wv.wav.spx.tta.oga.ape.aiff.opus").split(QSL("."));
    d->filetype=list.indexOf(extension);
    d->icon = QString(":/src/icons/%1.png").arg(extension);
}

void Tag::setTag1(const int tagID,const QString &value)
{
    switch (tagID) {
    case COMPOSER:    d->composer = value; break;
    case ALBUM:       d->album=value; break;
    case TITLE:       d->title=value; break;
    case PERFORMER:   d->performer=value; break;
    case ARTIST:      d->artist=value; break;
    case GENRE:       d->genre=value; break;
    case YEAR:        d->year=value; break;
    case COMMENT:     d->comment=value; break;
    case TRACKNUMBER: d->tracknumber=value; break;
    case TOTALTRACKS: d->totalTracks=value; break;
    case CONDUCTOR:   d->conductor=value; break;
    case ORCHESTRA:   d->orchestra=value; break;
    case SUBTITLE:    d->subtitle=value; break;
    case KEY:         d->key=value; break;
    case ALBUMARTIST: d->albumArtist=value; break;

    case CATEGORY :   d->category=value; break;
    case PUBLISHER :  d->publisher=value; break;
    case COPYRIGHT :  d->copyright=value; break;
    case MOOD :       d->mood=value; break;
    case TEMPO :      d->tempo=value; break;
    case LYRICIST :   d->lyricist=value; break;
    case LYRICS :     d->lyrics=value; break;
    case DISCNUMBER : d->discnumber=value; break;
    case TOTALDISCS : d->totaldiscs=value; break;
    case ENCODEDBY :  d->encodedby=value; break;
    case REMIXEDBY :  d->remixedby=value; break;
    case RATING :     d->rating=value; break;
    case ORIGINALALBUM: d->originalalbum=value; break;
    case ORIGINALARTIST: d->originalartist=value; break;
    case ORIGINALLYRICIST: d->originallyricist=value; break;
    case URL :        d->url=value; break;
    case ISRC :       d->isrc=value; break;
    default: break;
    }
}

void Tag::setTag(QString &tag, const QString &newTag)
{
    if (tag==newTag) return;
    d->wasChanged = true;
    tag = newTag;
}

void Tag::setTag(const int tagID,const QString &value)
{
    switch (tagID) {
    case COMPOSER:    setTag(d->composer,value); break;
    case ALBUM:       setTag(d->album,value); break;
    case TITLE:       setTag(d->title,value); break;
    case PERFORMER:   setTag(d->performer,value); break;
    case ARTIST:      setTag(d->artist,value); break;
    case GENRE:       setTag(d->genre,value); break;
    case YEAR:        setTag(d->year,value); break;
    case COMMENT:     setTag(d->comment,value); break;
    case TRACKNUMBER: setTag(d->tracknumber,value); break;
    case TOTALTRACKS: setTag(d->totalTracks,value); break;
    case CONDUCTOR:   setTag(d->conductor,value); break;
    case ORCHESTRA:   setTag(d->orchestra,value); break;
    case SUBTITLE:    setTag(d->subtitle,value); break;
    case KEY:         setTag(d->key,value); break;
    case ALBUMARTIST: setTag(d->albumArtist,value); break;

    case CATEGORY :   setTag(d->category,value); break;
    case PUBLISHER :  setTag(d->publisher,value); break;
    case COPYRIGHT :  setTag(d->copyright,value); break;
    case MOOD :       setTag(d->mood,value); break;
    case TEMPO :      setTag(d->tempo,value); break;
    case LYRICIST :   setTag(d->lyricist,value); break;
    case LYRICS :     setTag(d->lyrics,value); break;
    case DISCNUMBER : setTag(d->discnumber,value); break;
    case TOTALDISCS : setTag(d->totaldiscs,value); break;
    case ENCODEDBY :  setTag(d->encodedby,value); break;
    case REMIXEDBY :  setTag(d->remixedby,value); break;
    case RATING :     setTag(d->rating,value); break;
    case ORIGINALALBUM: setTag(d->originalalbum,value); break;
    case ORIGINALARTIST: setTag(d->originalartist,value); break;
    case ORIGINALLYRICIST: setTag(d->originallyricist,value); break;
    case URL :        setTag(d->url,value); break;
    case ISRC :       setTag(d->isrc,value); break;
    default: break;
    }
}


void Tag::setUserTag(const QString &id, const QString &s)
{
    if (d->otherTags.value(id)==s) return;
    d->wasChanged=true;
    d->otherTags.insert(id,s);
}

/* ================================================*/
CoverImage Tag::image() const
{
    return d->image;
}

void Tag::setImage(const CoverImage &image, bool updateStatus)
{
    if (d->image==image) return;
    if (updateStatus)  d->wasChanged = true;
    d->image = image;
}

void Tag::setImagePixmap(const QByteArray &pixmap, bool updateStatus)
{
    if (d->image.pixmap() == pixmap) return;
    if (updateStatus) d->wasChanged = true;
    d->image.setPixmap(pixmap);
}

void Tag::setImageDescription(const QString &description, bool updateStatus)
{
    if (d->image.description() == description) return;
    if (updateStatus) d->wasChanged = true;
    d->image.setDescription(description);
}

void Tag::setImageMimetype(const QString &mimetype, bool updateStatus)
{
    if (d->image.mimetype() == mimetype) return;
    if (updateStatus) d->wasChanged = true;
    d->image.setMimetype(mimetype);
}

void Tag::setImageType(int type, bool updateStatus)
{
    if (d->image.type() == type) return;
    if (updateStatus) d->wasChanged = true;
    d->image.setType(type);
}

void Tag::removeImage(bool updateStatus)
{
    if (d->image.isEmpty()) return;
    if (updateStatus) d->wasChanged = true;
    d->image.clear();
}

/* =========================================*/
bool Tag::operator==(const Tag &t) const
{
    for (int i=0; i<TAGSCOUNT; ++i)
        if (tag(i)!=t.tag(i))
            return false;
    return (d->image == t.d->image
            && d->wasChanged == t.d->wasChanged
            && d->otherTags == t.d->otherTags
            && d->fullFileName == t.d->fullFileName);
}

void Tag::clear()
{
    removeAllTags();
    d->replayGainInfo.clear();
    d->image.clear();
    d->length=0;
    d->bitrate.clear();
    d->sampleRate=0;
    d->channels=0;
    d->wasChanged=false;
}

void Tag::setChanged(const bool changed)
{
    d->wasChanged = changed;
}

void Tag::removeReplayGainInfo(bool updateStatus)
{
    if (d->replayGainInfo.isEmpty()) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.clear();
}

void Tag::setReplayGainInfo(const ReplayGainInfo &rg, bool updateStatus)
{
    if (d->replayGainInfo == rg) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo = rg;
}

void Tag::setLength(int length)
{
    d->length=length;
}

void Tag::appendTag(const int id, const QString &value)
{
    switch (id) {
    case COMPOSER:    {append(d->composer,value); break;}
    case ALBUM:       {append(d->album,value); break;}
    case TITLE:       {append(d->title,value); break;}
    case PERFORMER:   {append(d->performer,value); break;}
    case ARTIST:      {append(d->artist,value); break;}
    case GENRE:       {append(d->genre,value); break;}
    case YEAR:        {append(d->year,value); break;}
    case COMMENT:     {append(d->comment,value); break;}
    case TRACKNUMBER: {append(d->tracknumber,value); break;}
    case TOTALTRACKS: {append(d->totalTracks,value); break;}
    case CONDUCTOR:   {append(d->conductor,value); break;}
    case ORCHESTRA:   {append(d->orchestra,value); break;}
    case SUBTITLE:    {append(d->subtitle,value); break;}
    case KEY:         {append(d->key,value); break;}
    case ALBUMARTIST: {append(d->albumArtist,value); break;}
    case CATEGORY :   {append(d->category,value); break;}
    case PUBLISHER :  {append(d->publisher,value); break;}
    case COPYRIGHT :  {append(d->copyright,value); break;}
    case MOOD :       {append(d->mood,value); break;}
    case TEMPO :      {append(d->tempo,value); break;}
    case LYRICIST :   {append(d->lyricist,value); break;}
    case LYRICS :     {append(d->lyrics,value); break;}
    case DISCNUMBER : {append(d->discnumber,value); break;}
    case TOTALDISCS : {append(d->totaldiscs,value); break;}
    case ENCODEDBY :  {append(d->encodedby,value); break;}
    case REMIXEDBY :  {append(d->remixedby,value); break;}
    case RATING :     {append(d->rating,value); break;}
    case ORIGINALALBUM: {append(d->originalalbum,value); break;}
    case ORIGINALARTIST: {append(d->originalartist,value); break;}
    case ORIGINALLYRICIST: {append(d->originallyricist,value); break;}
    case URL :        {append(d->url,value); break;}
    case ISRC :       {append(d->isrc,value); break;}
    default: break;
    }
}

void Tag::appendUserTag(const QString &id, const QString &value)
{
    if (!d->otherTags.contains(id)) {
        d->otherTags.insert(id,value);
        return;
    }
    QStringList v = d->otherTags.value(id).split(QSL(";"));
    if (!v.contains(value)) {
        v << value;
        d->otherTags.insert(id,v.join(QSL(";")));
    }
}

void Tag::setAlbumPeak(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.albumPeak==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.albumPeak=value;
}

void Tag::setAlbumGain(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.albumGain==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.albumGain = value;
}

void Tag::setTrackPeak(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.trackPeak==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.trackPeak = value;
}

void Tag::setTrackGain(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.trackGain==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.trackGain = value;
}

void Tag::setAlbumMinMax(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.albumMinMax==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.albumMinMax = value;
}

void Tag::setTrackMinMax(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.trackMinMax==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.trackMinMax = value;
}

void Tag::setLoudness(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.loudness==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.loudness = value;
}

void Tag::setReplayGainUndo(const QString &value, bool updateStatus)
{
    if (d->replayGainInfo.undo==value) return;
    if (updateStatus) d->wasChanged = true;
    d->replayGainInfo.undo = value;
}


void Tag::setUserTags1(const QMap<QString, QString> &other)
{
    d->otherTags = other;
}

void Tag::removeStandardTags(bool /*updateStatus*/)
{
    d->album.clear(); //a
    d->artist.clear(); //A
    d->title.clear(); //t
    d->performer.clear(); //p
    d->comment.clear(); //C
    d->genre.clear(); //g
    d->composer.clear(); //c
    d->year.clear(); //y
    d->tracknumber.clear(); //n
    d->totalTracks.clear(); //T
    d->conductor.clear(); //d
    d->orchestra.clear(); //o
    d->subtitle.clear(); //s
    d->key.clear(); //k
    d->albumArtist.clear(); //u
    d->category.clear(); //G
    d->publisher.clear(); //P
    d->copyright.clear(); //X
    d->mood.clear(); //m
    d->tempo.clear(); //B
    d->lyricist.clear(); //w
    d->lyrics.clear(); //W
    d->discnumber.clear(); //D
    d->totaldiscs.clear(); //S
    d->encodedby.clear(); //e
    d->remixedby.clear(); //x
    d->rating.clear(); //R
    d->originalalbum.clear(); //O
    d->originalartist.clear(); //v
    d->originallyricist.clear(); //i
    d->url.clear(); //U
    d->isrc.clear(); //I
    d->wasChanged=true;
}

void Tag::removeAllTags(bool updateStatus)
{
    removeUserTags(updateStatus);
    removeStandardTags(updateStatus);
    removeImage(updateStatus);
    removeReplayGainInfo(updateStatus);
}

void Tag::removeUserTags(bool updateStatus)
{
    if (d->otherTags.isEmpty()) return;
    if (updateStatus) d->wasChanged = true;
    d->otherTags.clear();
}
